from typing import Dict, List, Union

Agg = Dict[str, str]  # dict of size 1: mapping colomn -> aggregation function


def add_aggregation_columns(
        df, *,
        group_cols: Union[str, List[str]],
        aggregations: Dict[str, Agg]
):
    """
    df =
       year  group libelle      mois  value  total
    0  2018  e-EKO       a  20180101      3      3
    1  2018  e-EKO       b  20180101      0      3
    2  2018  e-EKO       c  20180101      4      7
    3  2018    DOQ       q  20180101     33     33
    4  2018    DOQ       b  20180101     54     87
    5  2018     DO       q  20180101     13     13
    6  2018     DO       c  20180101     14     27

    df = groupby_append(
        df,
        ['mois', 'group'],
        {
            'max_total': {'total': 'max'},
            'sum_total': {'total': 'sum'},
            'mean_value': {'value': 'mean'}
        }
    )

    df =
       year  group libelle      mois  value  total  max_total  sum_total  mean_value
    0  2018  e-EKO       a  20180101      3      3          7         13    2.333333
    1  2018  e-EKO       b  20180101      0      3          7         13    2.333333
    2  2018  e-EKO       c  20180101      4      7          7         13    2.333333
    3  2018    DOQ       q  20180101     33     33         87        120   43.500000
    4  2018    DOQ       b  20180101     54     87         87        120   43.500000
    5  2018     DO       q  20180101     13     13         27         40   13.500000
    6  2018     DO       c  20180101     14     27         27         40   13.500000
    """
    group = df.groupby(group_cols)
    for new_col, aggs in aggregations.items():
        assert len(aggs) == 1
        (col, agg), *_ = aggs.items()
        df[new_col] = group[col].transform(agg)
    return df



import os
import shutil
import tempfile

import joblib
import pytest
from requests import HTTPError

from tests.tools import DF, DF2
from toucan_data_sdk.sdk import ToucanDataSdk, InvalidQueryError


def gen_client(mocker):
    class Response:
        content = 10

        def raise_for_status(self):
            pass

    resp = Response()
    client = mocker.MagicMock()
    client.sdk.post.return_value = resp
    return client


@pytest.fixture(name='sdk', scope='function')
def gen_data_sdk(mocker):
    sdk = ToucanDataSdk(
        instance_url='https://api-myinstance.toucantoco.com',
        small_app='demo',
        auth=('', '')
    )
    sdk.client = gen_client(mocker)
    yield sdk
    if os.path.exists(sdk.EXTRACTION_CACHE_PATH):
        shutil.rmtree(sdk.EXTRACTION_CACHE_PATH)


@pytest.fixture(name='sdk_old', scope='function')
def gen_data_sdk_old(mocker):
    sdk_old = ToucanDataSdk(
        instance_url='https://api-myinstance.toucantoco.com/demo',
        auth=('', '')
    )
    sdk_old.client = gen_client(mocker)
    yield sdk_old
    if os.path.exists(sdk_old.EXTRACTION_CACHE_PATH):
        shutil.rmtree(sdk_old.EXTRACTION_CACHE_PATH)


def gen_client_error(mocker):
    class Response:
        content = 10

        def raise_for_status(self):
            raise HTTPError()

    resp = Response()
    client = mocker.MagicMock()
    client.sdk.post.return_value = resp
    return client


@pytest.fixture(name='sdk_error', scope='function')
def gen_data_sdk_error(mocker):
    sdk = ToucanDataSdk('some_url', small_app='demo', auth=('', ''))
    sdk.client = gen_client_error(mocker)
    yield sdk
    if os.path.exists(sdk.EXTRACTION_CACHE_PATH):
        shutil.rmtree(sdk.EXTRACTION_CACHE_PATH)


@pytest.fixture(name='tmp_dir', scope='module')
def gen_tmp_dir():
    return tempfile.gettempdir()


@pytest.fixture(name='tmp_file', scope='function')
def gen_tmp_file(tmp_dir):
    tmp_file = tempfile.NamedTemporaryFile(dir=tmp_dir)
    yield tmp_file
    tmp_file.close()


@pytest.fixture(name='tmp_file2', scope='function')
def gen_tmp_file2(tmp_dir):
    tmp_file = tempfile.NamedTemporaryFile(dir=tmp_dir)
    yield tmp_file
    tmp_file.close()


def test_datasources(sdk, mocker):
    """It should use the cache properly"""
    mock_cache_exists = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.cache_exists')
    mock_read_cache = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.read_from_cache')
    mock_read_sdk = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.read_datasources_from_sdk')
    # 1. Cache directory exists
    mock_cache_exists.return_value = True
    mock_read_cache.return_value = {"domain_1": 1}
    assert sdk.get_dfs() == {"domain_1": 1}

    # 2. Cache directory does not exist
    mock_read_cache.reset_mock()
    mock_cache_exists.return_value = False
    mock_read_sdk.return_value = {"domain_2": 1}
    assert sdk.get_dfs() == {"domain_2": 1}
    mock_read_cache.assert_not_called()

    # 3. Cache directory exists with one domain
    mock_cache_exists.return_value = True
    mock_read_cache.return_value = {"domain_1": 1}
    assert sdk.get_dfs(['a']) == {"domain_1": 1}

    # 4. Cache directory doesn't exists with one domain
    mock_read_cache.reset_mock()
    mock_cache_exists.return_value = False
    mock_read_sdk.return_value = {"domain_2": 1}
    assert sdk.get_dfs(['a']) == {"domain_2": 1}
    mock_read_cache.assert_not_called()
    assert sdk.small_app_url == 'https://api-myinstance.toucantoco.com/demo'


def test_dfs_complex(sdk, mocker):
    """It should use the cache properly"""
    mock_cache_exists = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.cache_exists')
    mock_read_cache = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.read_from_cache')
    mock_read_sdk = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.read_datasources_from_sdk')

    mock_cache_exists.side_effect = [True, False]
    mock_read_cache.return_value = {"domain_1": 1}
    mock_read_sdk.return_value = {"domain_2": 1}
    assert sdk.get_dfs(['a', 'b']) == {"domain_1": 1, "domain_2": 1}


def test_dfs_http_error(sdk_error):
    """It should use the cache properly"""
    with pytest.raises(HTTPError):
        sdk_error.get_dfs()


def test_read_from_cache(sdk):
    with tempfile.TemporaryDirectory() as tmp_dir:
        extraction_dir = os.path.join(tmp_dir, sdk.EXTRACTION_CACHE_PATH)
        sdk.EXTRACTION_CACHE_PATH = extraction_dir
        os.makedirs(extraction_dir)

        joblib.dump(DF, os.path.join(extraction_dir, 'a'))
        joblib.dump(DF2, os.path.join(extraction_dir, 'b'))

        dfs = sdk.read_from_cache()
        assert 'a' in dfs
        assert 'b' in dfs
        assert DF.equals(dfs['a'])
        assert DF2.equals(dfs['b'])

        dfs = sdk.read_from_cache(['a'])
        assert 'a' in dfs
        assert 'b' not in dfs
        assert DF.equals(dfs['a'])


def test_read_from_sdk(sdk, mocker):
    mock_extract = mocker.patch('toucan_data_sdk.sdk.extract')
    mock_extract.return_value = {'df': DF, 'df2': DF2}
    dfs = sdk.read_datasources_from_sdk()
    assert dfs == {'df': DF, 'df2': DF2}


def test_write(sdk, mocker):
    mock_extract = mocker.patch('toucan_data_sdk.sdk.extract')
    mock_extract.return_value = {'a': DF, 'b': DF2}

    with tempfile.TemporaryDirectory() as tmp_dir:
        extraction_dir = os.path.join(tmp_dir, sdk.EXTRACTION_CACHE_PATH)
        sdk.EXTRACTION_CACHE_PATH = extraction_dir

        sdk.write({'a': DF, 'b': DF2})

        assert os.path.exists(extraction_dir)
        assert 'a' in os.listdir(extraction_dir)
        assert 'b' in os.listdir(extraction_dir)


def test_invalidate_cache(sdk):
    with tempfile.TemporaryDirectory() as tmp_dir:
        extraction_dir = os.path.join(tmp_dir, sdk.EXTRACTION_CACHE_PATH)
        sdk.EXTRACTION_CACHE_PATH = extraction_dir
        os.makedirs(extraction_dir)
        joblib.dump(DF, os.path.join(extraction_dir, 'a'))
        joblib.dump(DF2, os.path.join(extraction_dir, 'b'))

        assert sdk.cache_exists()
        sdk.invalidate_cache()
        assert not sdk.cache_exists()

        os.makedirs(extraction_dir)
        joblib.dump(DF, os.path.join(extraction_dir, 'a'))
        joblib.dump(DF2, os.path.join(extraction_dir, 'b'))

        assert sdk.cache_exists('a')
        sdk.invalidate_cache(['a'])
        assert not sdk.cache_exists('a')
        assert sdk.cache_exists('b')
        assert sdk.cache_exists()


def test_invalidate_cache_exception(sdk):
    with tempfile.TemporaryDirectory() as tmp_dir:
        sdk.invalidate_cache()

        extraction_dir = os.path.join(tmp_dir, sdk.EXTRACTION_CACHE_PATH)
        sdk.EXTRACTION_CACHE_PATH = extraction_dir
        os.makedirs(extraction_dir)
        joblib.dump(DF, os.path.join(extraction_dir, 'a'))

        assert sdk.cache_exists()
        sdk.invalidate_cache(['b', 'a'])
        assert not sdk.cache_exists('a')


def test_augment(sdk):
    sdk.client.config.augment.get().text = 'yo'
    assert sdk.get_augment() == 'yo'


def test_etl(sdk):
    sdk.client.config.etl.get().json.return_value = {'yo': 'del'}
    assert sdk.get_etl() == {'yo': 'del'}


def test_basemaps(sdk):
    sdk.client.basemaps.post().json.return_value = {'lala': 'lo'}
    assert sdk.query_basemaps({}) == {'lala': 'lo'}
    with pytest.raises(InvalidQueryError):
        sdk.query_basemaps('yo')


def test_sdk_compatibility(sdk_old, mocker):
    """It should use the cache properly"""
    mock_cache_exists = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.cache_exists')
    mock_read_cache = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.read_from_cache')
    mocker.patch('toucan_data_sdk.sdk.ToucanDataSdk.read_datasources_from_sdk')

    # 1. Cache directory exists
    mock_cache_exists.return_value = True
    mock_read_cache.return_value = {"domain_1": 1}
    assert sdk_old.get_dfs() == {"domain_1": 1}
    assert sdk_old.small_app_url == 'https://api-myinstance.toucantoco.com/demo'


def test_get_domain(sdk):
    sdk.client.output_domain['my_domain'].post().json.return_value = {
        'result': [{'_id': {'$oid': '5b449af2291ebbd9087f6260'}, 'toto': 2010,
                    'label': 'Maladie', 'value': 4.2, 'domain': '0_201_1'},
                   {'_id': {'$oid': '5b449af2291ebbd9087f6261'}, 'toto': 2010,
                    'label': 'Accident Travail', 'value': 1.2, 'domain': '0_201_1'},
                   {'_id': {'$oid': '5b449af2291ebbd9087f6262'}, 'toto': 2010,
                    'label': 'Invalidité', 'value': 1.3, 'domain': '0_201_1'}],
        'lastDocId': 'ab31cd'
    }
    sdk.client.output_domain['my_domain']['ab31cd'].post().json.return_value = {
        'result': [{'_id': {'$oid': '5b449af2291ebbd9087f6260'}, 'toto': 2011,
                    'label': 'Maladie', 'value': 4.1, 'domain': '0_201_1'},
                   {'_id': {'$oid': '5b449af2291ebbd9087f6261'}, 'toto': 2011,
                    'label': 'Invalidité', 'value': 3.7, 'domain': '0_201_1'}],
        'lastDocId': None
    }
    dfs = sdk.get_domains(['my_domain'])
    assert set(dfs['my_domain']) == {'toto', 'label', 'value', 'domain'}
    assert dfs['my_domain'].shape == (5, 4)


def test_get_domains(sdk):
    sdk.client.output_domain['bla'].post().json.side_effect = [
        {
            'result': [{'_id': {'$oid': '5b449af2291ebbd9087f6260'}, 'toto': 2010,
                        'domain': '0_201_1'}],
            'lastDocId': 'ab31cd'
        },
        {
            'result': [
                {'_id': {'$oid': '5b449af2291ebbd9087f6260'}, 'tutu': 2011,
                 'domain': '0_201_2'}],
            'lastDocId': None
        }
    ]
    sdk.client.output_domain['bla']['ab31cd'].post().json.return_value = {
        'result': [{'_id': {'$oid': '5b449af2291ebbd9087f6260'}, 'titi': 2011,
                    'domain': '0_201_1'}],
        'lastDocId': None
    }
    sdk.client.metadata.get().json.return_value = [
        {'domain': 'a_domain'}, {'domain': 'b_domain'}
    ]
    dfs = sdk.get_domains()
    assert set(dfs['a_domain']) == {'toto', 'titi', 'domain'}
    assert set(dfs['b_domain']) == {'tutu', 'domain'}
    assert dfs['a_domain'].shape == (2, 3)
    assert dfs['b_domain'].shape == (1, 2)


def test_domain_cache(mocker, sdk):
    mock_cache_exists = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.cache_exists')
    mock_read_cache = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.read_from_cache')
    mock_cache_exists.return_value = True
    mock_read_cache.return_value = {"domain_1": 1}
    assert sdk.get_domains('domain_1') == {"domain_1": 1}


def test_traceback(sdk):
    class Response:
        def __init__(self, ok):
            self.ok = ok
            filename = './tests/fixtures/deleteme-exception.dump'
            with open(filename, 'rb') as f:
                self.content = f.read()

        def json(self):
            return {'test': "yo"}

    sdk.client.tracebacks.latest.get.return_value = Response(True)
    tb_values = sdk.load_latest_traceback()
    assert 'linechart' in tb_values
    assert tb_values['VAR'] == 1

    sdk.client.tracebacks.latest.get.return_value = Response(False)
    tb = sdk.load_latest_traceback()
    assert {'test': "yo"} == tb



import joblib
import types
import io
import os


def _from_pickled_dict(d):
    def from_picklable(obj):
        if obj is None:
            return None
        try:
            f = io.BytesIO(obj)
            return joblib.load(f)
        except Exception:
            return None
    unpickled = {k: from_picklable(v) for k, v in d.items()}
    unpickling_failed = {k: v for k, v in unpickled.items() if v is None}
    unpickled = {k: v for k, v in unpickled.items() if v is not None}
    unpickled['__unpickling_failed__'] = list(unpickling_failed.keys())
    return unpickled


def _from_serializable_traceback(d):
    tb = d['tb']
    tb = types.SimpleNamespace(**tb)
    tb.tb_frame = types.SimpleNamespace(**tb.tb_frame)
    tb.tb_frame.f_code = types.SimpleNamespace(**tb.tb_frame.f_code)
    tb.tb_frame.f_locals = _from_pickled_dict(tb.tb_frame.f_locals)
    tb.tb_frame.f_globals = _from_pickled_dict(tb.tb_frame.f_globals)
    return d['exc_type'], d['exc_value'], tb


def _print_tb(exc_value, tb):
    lineno = tb.tb_lineno
    lines = tb.sourcecode.splitlines()

    relative_err_lineno = min(11, lineno)
    first_lineno = max(0, lineno - 12)
    last_lineno = min(len(lines), lineno + 4)

    lines = lines[first_lineno:last_lineno]
    lines_numbers = list(range(first_lineno + 1, last_lineno + 1))
    lines_numbers = [str(n).rjust(4) + '|' for n in lines_numbers]
    lines_numbers[relative_err_lineno] = '====>'

    for lineno, line in zip(lines_numbers, lines):
        print(lineno + line)
    print('−−−−−−−−−−−−')
    print(exc_value)


def _inject_into_globals(tb):
    tb_values = {}
    for k, v in tb.tb_frame.f_globals.items():
        if k not in globals():
            tb_values[k] = v
    for k, v in tb.tb_frame.f_locals.items():
        if k not in globals():
            tb_values[k] = v
    return tb_values


def load_traceback(filename):
    stb = joblib.load(filename)
    os.remove(filename)
    exc_type, exc_value, tb = _from_serializable_traceback(stb)
    _print_tb(exc_value, tb)
    return _inject_into_globals(tb)



import pandas as pd


def date_requester_generator(start_date, end_date, frequency,
                             format='%Y-%m-%d', granularities=None,
                             others_format=None, times_delta=None):
    """
    Return a DataFrame containing at least 3 columns :
    - "DATE" : Label of date
    - "DATETIME" : Date in datetime dtype
    - "GRANULARITY" : Granualrity of date

    Arguments
    #########

    Mandatory :
    -----------
    - start_date (str) : start date in %Y-%m%d format
    - end_date (str): end date in %Y-%m%d format
    - frequency (str) : http://pandas.pydata.org/pandas-docs/stable/timeseries.html#offset-aliases

    Optional :
    ----------
    - format: format of the date. !!! only use if granularity is None
        NB : same format in Toucan Toco
        Example : '%d/%m/%Y'
        https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior
    - granularities (dict):
        - keys : name of the granularity
        - values (str): Format of the granularity.
            NB : same format in Toucan Toco
            Example : '%d/%m/%Y'
            https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior
    - others_format (dict) : Add new columns for each key
        - key (str) : name of the column
        - values (str): format of the granularity.
            NB : same format in Toucan Toco
            Example : '%d/%m/%Y'
            https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior
    - times_delta (dict) : Add new columns for each key
        - key (str) : name of the column
        - values (str): time delta
            Examples : '+1 day' '+3 day' '-4 month'
    """

    granularities = granularities or {'date': format}
    others_format = others_format or {}
    times_delta = times_delta or {}

    # Base DataFrame
    columns_list = ['DATE', 'DATETIME', 'GRANULARITY', *others_format, *times_delta]
    result_df = {col_name: [] for col_name in columns_list}

    # Generate the range
    date_range = pd.date_range(start=start_date, end=end_date, freq=frequency)

    for granularity_name, granularity_format in granularities.items():
        date_range_label = date_range.strftime(granularity_format)
        a = list(set(date_range_label))
        index_unique = list(set([a.index(x) for x in date_range_label]))
        date_range_datetime = date_range[index_unique]
        date_range_label = date_range_label.unique()

        result_df['DATE'] += list(date_range_label)
        result_df['DATETIME'] += list(date_range_datetime)
        result_df['GRANULARITY'] += [granularity_name] * len(date_range_label)

        for col_name, other_format in others_format.items():
            result_df[col_name] += list(date_range_datetime.strftime(other_format))

        for col_name, time_delta in times_delta.items():
            result_df[col_name] += list((date_range_datetime + pd.Timedelta(time_delta))
                                        .strftime(granularity_format))

    return pd.DataFrame(result_df)



def sort(df, columns, order='asc'):
    """
    Sort DataFrame
    :param df: DataFrame
    :param columns: name of the columns to sort
    :param order: asc or desc
    :return: DataFrame
    """
    ascending = order != 'desc'
    return df.sort_values(columns, ascending=ascending)



"""date filtering helpers."""

from datetime import date, datetime
import re
from uuid import uuid4

import pandas as pd


def _norm_date(datestr: str, date_fmt: str) -> date:
    """normalize symbolic date values (e.g. 'TODAY')

    Convert a symbolic value in a valid date.
    Currenlty known symbolic values are 'TODAY', 'YESTERDAY' and 'TOMORROW'.

    NOTE: This function will return `date` (not `datetime`) instances.

    Parameters:
        `datestr`: the date to parse, formatted as `date_fmt`
        `date_fmt`: expected output date format

    Returns:
        The interpreted date as a datetime.datetime object.
        If `datestr` doesn't match any of the known symbolic names, it just parses it.

    """
    try:
        days = {'TODAY': 0, 'YESTERDAY': -1, 'TOMORROW': 1}[datestr.upper()]
        return date.today() + pd.Timedelta(days=days)
    except KeyError:
        return datetime.strptime(datestr, date_fmt).date()


def parse_date(datestr: str, date_fmt: str) -> date:
    """parse `datestr` and return corresponding date object.

    `datestr` should be a string matching `date_fmt` and parseable by
    `strptime` but some offset can also be added using `(datestr) + OFFSET` or
    `(datestr) - OFFSET` syntax. When using this syntax, `OFFSET` should be
    understable by `pandas.Timedelta` (cf.
    http://pandas.pydata.org/pandas-docs/stable/timedeltas.html) and `datestr`
    MUST be wrapped with parenthesis.

    Additionally, the following symbolic names are supported:
    `TODAY`, `YESTERDAY`, `TOMORROW`.

    Example usage:

    >>> parse_date('2018-01-01', '%Y-%m-%d')
    datetime.date(2018, 1, 1)
    >>> parse_date('(2018-01-01) + 1day', '%Y-%m-%d')
    datetime.date(2018, 1, 2)

    Parameters:
        `datestr`: the date to parse, formatted as `date_fmt`
        `date_fmt`: expected date format

    Returns:
        The `date` object. If date could not be parsed, a ValueError
        will be raised.
    """
    rgx = re.compile(r'\((?P<date>.*)\)(\s*(?P<sign>[+-])(?P<offset>.*))?$')
    datestr = datestr.strip()
    match = rgx.match(datestr)
    # if regexp doesn't match, date must match the expected format
    if match is None:
        return _norm_date(datestr, date_fmt)
    datestr = match.group('date').strip()
    dateobj = _norm_date(datestr, date_fmt)
    offset = match.group('offset')
    if offset:
        sign = match.group('sign')
        delta = pd.Timedelta(offset)
        dateobj = (dateobj + delta) if sign == '+' else (dateobj - delta)
    return dateobj


def filter_by_date(df, date_col, date_fmt='%Y-%m-%d', start=None, stop=None, atdate=None):
    """filter dataframe `df` by date.

    This function will interpret `start`, `stop` and `atdate` and build
    the corresponding date range. The caller must specify either:

    - `atdate`: keep all rows matching this date exactly,
    - `start`: keep all rows matching this date onwards.
    - `stop`: keep all rows matching dates before this one.
    - `start` and `stop`: keep all rows between `start` and `stop`,

    Any other combination will raise an error. The lower bound of the date range
    will be included, the upper bound will be excluded.

    When specified, `start`, `stop` and `atdate` values are expected to match the
    `date_fmt` format or a known symbolic value (i.e. 'TODAY', 'YESTERDAY' or
    'TOMORROW').

    Additionally, the offset syntax "(date) + offset" is also supported (Mind
    the parenthesis around the date string). In that case, the offset must be
    one of the syntax supported by `pandas.Timedelta` (cf.
    http://pandas.pydata.org/pandas-docs/stable/timedeltas.html)

    Parameters:
        `df`: the dataframe to filter
        `date_col`: the name of the dataframe's column to filter on
        `date_fmt`: expected date format in column `date_col`
        `start`: if specified, lower bound (included) of the date range
        `stop`: if specified, upper bound (excluded) of the date range
        `atdate`: if specified, the exact date we're filtering on

    Returns:
        The filtered dataframe
    """
    mask = None
    if start is None and stop is None and atdate is None:
        raise TypeError('either "start", "stop" or "atdate" must be specified')
    if start is not None and atdate is not None:
        raise TypeError('"start" and "atdate" are mutually exclusive')
    if stop is not None and atdate is not None:
        raise TypeError('"stop" and "atdate" are mutually exclusive')
    # add a new column that will hold actual date objects instead of strings.
    # This column is just temporary and will be removed before returning the
    # filtered dataframe.
    filtercol = str(uuid4())
    df[filtercol] = pd.to_datetime(df[date_col], format=date_fmt)
    if atdate is not None:
        mask = df[filtercol] == parse_date(atdate, date_fmt)
    elif start is not None and stop is not None:
        mask = ((df[filtercol] >= parse_date(start, date_fmt)) &
                (df[filtercol] < parse_date(stop, date_fmt)))
    elif stop is None:
        mask = df[filtercol] >= parse_date(start, date_fmt)
    elif start is None:
        mask = df[filtercol] < parse_date(stop, date_fmt)
    return df[mask].drop(filtercol, axis=1)



import operator as _operator
from typing import List

MATH_CHARACTERS = '()+-/*%.'


def _basic_math_operation(df, new_column, column_1, column_2, op):
    """
    Basic mathematical operation to apply operator on `column_1` and `column_2`
    Both can be either a number or the name of a column of `df`
    Will create a new column named `new_column`
    """
    if not isinstance(column_1, (str, int, float)):
        raise TypeError(f'column_1 must be a string, an integer or a float')
    if not isinstance(column_2, (str, int, float)):
        raise TypeError(f'column_2 must be a string, an integer or a float')

    if isinstance(column_1, str):
        column_1 = df[column_1]
    if isinstance(column_2, str):
        column_2 = df[column_2]
    operator = getattr(_operator, op)
    df[new_column] = operator(column_1, column_2)
    return df


def add(df, new_column, column_1, column_2):
    """Add df[value] (value: 'str') or value (number) to column_1"""
    return _basic_math_operation(df, new_column, column_1, column_2, op='add')


def subtract(df, new_column, column_1, column_2):
    """Subtract df[value] (value: 'str') or value (number) to column_1"""
    return _basic_math_operation(df, new_column, column_1, column_2, op='sub')


def multiply(df, new_column, column_1, column_2):
    """Multiply df[value] (value: 'str') or value (number) and column_1"""
    return _basic_math_operation(df, new_column, column_1, column_2, op='mul')


def divide(df, new_column, column_1, column_2):
    """Divide df[value] (value: 'str') or value (number) to column_1"""
    return _basic_math_operation(df, new_column, column_1, column_2, op='truediv')


def is_float(x):
    try:
        float(x)
    except ValueError:
        return False
    else:
        return True


class Token(str):
    """
    A formula is a string like this '"2018  " - 2017 + (a - b)'
    In order to parse it, we split it in different tokens and keep track if it was
    quoted or not.
    E.g. in the formula above, `2017` is a number whereas `"2018"` is a column name.
    even though both are strings.
    """
    def __new__(cls, text, quoted=False):
        string = super().__new__(cls, text.strip())
        string.quoted = quoted
        return string


def _parse_formula(formula_str) -> List[Token]:
    tokens = []
    current_word = ''
    quote_to_match = None
    for x in formula_str:
        if x in ('"', "'") and not quote_to_match:
            quote_to_match = x
            continue
        if x == quote_to_match:
            tokens.append(Token(current_word, True))
            current_word = ''
            quote_to_match = None
            continue
        if quote_to_match or x not in MATH_CHARACTERS:
            current_word += x
        else:
            tokens.append(Token(current_word))
            current_word = ''
            tokens.append(Token(x))
    tokens.append(Token(current_word))
    if quote_to_match is not None:
        raise FormulaError('Missing closing quote in formula')
    return [t for t in tokens if t]


def formula(df, *, new_column, formula):
    """Compute math formula for df and put the result in `column`"""
    tokens = _parse_formula(formula)
    expression_splitted = []
    for t in tokens:
        # To use a column name with only digits, it has to be quoted!
        # Otherwise it is considered as a regular number
        if not t.quoted and (t in MATH_CHARACTERS or is_float(t)):
            expression_splitted.append(t)
        elif t in df.columns:
            expression_splitted.append(f'df["{t}"]')
        else:
            raise FormulaError(f'"{t}" is not a valid column name')
    expression = ''.join(expression_splitted)
    df[new_column] = eval(expression)
    return df


class FormulaError(Exception):
    """Raised when a formula is not valid"""



def rename(df, values=None, columns=None, locale=None):
    """
    Replaces data values and column names according to locale
    Args:
        df (pd.DataFrame): DataFrame to transform
        values (dict):
            - key (str): term to be replaced
            - value (dict):
                - key: locale
                - value: term's translation
        columns (dict):
            - key (str): columns name to be replaced
            - value (dict):
                - key: locale
                - value: column name's translation
        locale (str): locale
    """
    if values:
        to_replace = list(values.keys())
        value = [values[term][locale] for term in values]
        df = df.replace(to_replace=to_replace, value=value)
    if columns:
        _keys = list(columns.keys())
        _values = [column[locale] for column in columns.values()]
        columns = dict(list(zip(_keys, _values)))
        df = df.rename(columns=columns)
    return df



from typing import Dict, List, Union


def groupby(df, *, group_cols: Union[str, List[str]], aggregations: Dict[str, str]):
    """
    Aggregate values by groups.

    :param df: dataframe to group
    :param group_cols: list of columns used to group data
    :param aggregations: dictionnary of values columns to group as keys and aggregation function to
        use as values. Available aggregation functions:
         - 'sum'
         - 'mean'
         - 'median'
         - 'prod' (product)
         - 'std' (standard deviation)
         - 'var' (variance)
    :return: the summarized dataframe

    Example:

    Input df:

    | ENTITY | YEAR | VALUE_1 | VALUE_2 |
    |:------:|:----:|:-------:|:-------:|
    |    A   | 2017 |    10   |    3    |
    |    A   | 2017 |    20   |    1    |
    |    A   | 2018 |    10   |    5    |
    |    A   | 2018 |    30   |    4    |
    |    B   | 2017 |    60   |    4    |
    |    B   | 2017 |    40   |    3    |
    |    B   | 2018 |    50   |    7    |
    |    B   | 2018 |    60   |    6    |

    groupby(
        df=data,
        group_cols=['ENTITY', 'YEAR'],
        aggregations={
            'VALUE_1': 'sum',
            'VALUE_2': 'mean'
        }
    )

    returns:

    | ENTITY | YEAR | VALUE_1 | VALUE_2 |
    |:------:|:----:|:-------:|:-------:|
    |    A   | 2017 |    30   |   2.0   |
    |    A   | 2018 |    40   |   4.5   |
    |    B   | 2017 |   100   |   3.5   |
    |    B   | 2018 |   110   |   6.5   |

    """
    df = df.groupby(group_cols, as_index=False).agg(aggregations)
    return df



from toucan_data_sdk.utils.helpers import check_params_columns_duplicate


def compute_ffill_by_group(df, id_cols, reference_cols, value_col):
    """
    Compute ffill with groupby. There is a performance issue with a simple
    groupby/fillna (2017/07)
    - `id_cols` are the columns id to group,
    - `reference_cols` are the other columns used to order,
    - `value_col` is the name of the column to fill,

    Args:
        df (pd.DataFrame):
        id_cols (list(str)):
        reference_cols (list(str)):
        value_col (str):
    """
    check_params_columns_duplicate(id_cols + reference_cols + [value_col])
    df = df.sort_values(by=id_cols + reference_cols)
    df = df.set_index(id_cols)
    df['fill'] = 1 - df[value_col].isnull().astype(int)
    df['fill'] = df.groupby(
        level=list(range(0, len(id_cols) - 1))
    )['fill'].cumsum()
    df[value_col] = df[value_col].ffill()
    df.loc[df['fill'] == 0, value_col] = None
    del df['fill']
    return df.reset_index()



def drop_duplicates(df, columns):
    """
    Use only `columns` to identify duplicates and remove them.
    Set `columns` to None to use all of the columns
    """
    return df.drop_duplicates(columns)


def query_df(df, query):
    """
    Slice the data according to the provided query
    Basic usage like the one in the data query but in the postprocess
    Useful if you want to perform some slicing after a melt or pivot for example.
    Wired on the DataFrame.query method, see doc
    http://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.query.html#pandas.DataFrame.query
    Args:
        Query String
    """
    df = df.query(query)
    return df


query = query_df



def round_values(df, params):
    """
    This function replace numeric column values inplace by the rounded value.

    :param df: the dataframe to which we want to apply the round function
    :param kwargs: a dictionary with:
      - key: column name to which we want to apply the round function
      - value: Number of decimal places to round the column to
    :return: the df dataframe with the rounded value columns.


    Examples:

    Dataset df:

    ENTITY  VALUE_1  VALUE_2
       A     -1.5     -1.5
       A      0.4      0.4
       A       0        0
       A      1.6      1.6

    rank(df, params={'VALUE_1': None, 'VALUE_2': 1}) returns:

    ENTITY  VALUE_1  VALUE_2
       A     -2.0     -1.6
       A      0.0      0.4
       A      0.0      0.0
       A      2.0      1.6

    """

    for column, decimal in params.items():
        df[column] = df[column].round(decimal)
    return df



import copy
import tempfile

import pytest

from tests.tools import (
    default_zip_file,
    DF,
    DF2)
from toucan_data_sdk.sdk import extract_zip, extract, DataSdkError


@pytest.fixture(name='zip_content', scope='function')
def gen_zipfile():
    return copy.deepcopy(default_zip_file(DF, DF2))


@pytest.fixture(name='df', scope='function')
def gen_df():
    return DF.copy(deep=True)


@pytest.fixture(name='df2', scope='function')
def gen_df2():
    return DF2.copy(deep=True)


def test_extract_zip(zip_content):
    with tempfile.NamedTemporaryFile() as tmp_file:
        tmp_file.write(zip_content)
        tmp_file.flush()
        tmp_file.seek(0)

        dfs = extract_zip(tmp_file.name)
        assert 'df' in dfs
        assert 'df2' in dfs
        assert DF.equals(dfs['df'])
        assert DF2.equals(dfs['df2'])


def test_extract(df, df2, mocker):
    mock_extract_zip = mocker.patch('toucan_data_sdk.sdk.extract_zip')
    mock_extract_zip.return_value = 1

    # 1. Is a (valid) zip file
    zip_content = default_zip_file(df, df2)
    res = extract(zip_content)

    assert mock_extract_zip.call_count == 1
    assert res == 1

    # 2. Is not a zip file
    mock_is_zip_file = mocker.patch('zipfile.is_zipfile')
    mock_is_zip_file.return_value = False

    with pytest.raises(DataSdkError):
        extract(zip_content)

    # 3. Unknown input data
    with pytest.raises(DataSdkError):
        extract(b'string is not a valid input')



import pandas as pd


def convert_str_to_datetime(df, *, column=None, format=None):
    """
    Convert string column into datetime column
    :param df: Dataframe
    :param column: name of the column to format
    :param format: format of the values
    :return: df
    """
    df[column] = pd.to_datetime(df[column], format=format)
    return df


def convert_datetime_to_str(df, *, column=None, format=None, new_column=None):
    """
    Convert datetime column into string column
    :param df: Dataframe
    :param column: name of the column to format
    :param format: format of the result values
    :param new_column: name of the output
    :return: df
    """
    new_column = new_column or column
    df[new_column] = df[column].dt.strftime(format)
    return df


def change_date_format(df, *, column, output_format, input_format=None, new_column=None,
                       new_time_zone=None):
    """
    Convert datetime column into string column
    :param df: Dataframe
    :param column: name of the column to format
    :param output_format: format of the output values
    :param input_format: format of the input values (If None, let the parser detect it)
    :param new_column: name of the output column
    :param new_time_zone: name of new time zone (If None, no time zone conversion is done)
    :return: df
    """
    new_column = new_column or column
    df[new_column] = (pd.to_datetime(df[column], format=input_format, utc=True)
                      .dt.tz_convert(new_time_zone)
                      .dt.strftime(output_format))
    return df


def cast(df, column, type, new_column=None):
    """
    Convert column's type into type
    :param df: Dataframe
    :param column: name of the column to format
    :param type: desired type of the column
    :return: df
    """
    new_column = new_column or column
    df[new_column] = df[column].astype(type)
    return df



import numpy as np

__all__ = (
    'lower',
    'upper',
    'title',
    'capitalize',
    'swapcase',
    'length',
    'isalnum',
    'isalpha',
    'isdigit',
    'isspace',
    'islower',
    'isupper',
    'istitle',
    'isnumeric',
    'isdecimal',
    'strip',
    'lstrip',
    'rstrip',
    'center',
    'ljust',
    'rjust',
    'split',
    'rsplit',
    'partition',
    'rpartition',
    'find',
    'rfind',
    'index',
    'rindex',
    'startswith',
    'endswith',
    'concat',
    'contains',
    'repeat',
    'replace_pattern',
    # 'slice',
    # 'slice_replace',
    # 'count'
)


###################################################################################################
#                              METHODS WITH NO EXTRA PARAMETERS
#
# All these functions have the same signature:
# :param df: the dataframe
# :param column: the column
# :param new_column: the destination column (if not set, `column` will be used)
# :return: the transformed dataframe
###################################################################################################

def _generate_basic_str_postprocess(method_name, docstring):
    def f(df, column, new_column=None):
        method = getattr(df[column].str, method_name)
        new_column = new_column or column
        df.loc[:, new_column] = method()
        return df

    f.__name__ = method_name
    f.__doc__ = f'{docstring}\n' \
                f':param df: the dataframe\n' \
                f':param column: the column\n' \
                f':param new_column: the destination column (if not set, `column` will be used)\n' \
                f':return: the transformed dataframe'
    return f


doc = 'Compute length of each string of `column`'
length = _generate_basic_str_postprocess('len', doc)

# lower, upper, capitalize, title, swapcase
###################################################################################################
doc = 'Converts all characters of `column` to lowercase.'
lower = _generate_basic_str_postprocess('lower', doc)

doc = 'Converts all characters of `column` to uppercase.'
upper = _generate_basic_str_postprocess('upper', doc)

doc = 'Converts first character to uppercase and remaining ' \
      'to lowercase for each line of `column`.'
capitalize = _generate_basic_str_postprocess('capitalize', doc)

doc = 'Converts first character to uppercase and remaining ' \
      'to lowercase for each word of each line of `column`.'
title = _generate_basic_str_postprocess('title', doc)

doc = 'Converts uppercase to lowercase and lowercase to uppercase for each word of `column`.'
swapcase = _generate_basic_str_postprocess('swapcase', doc)

# isalnum, isalpha, isdigit, isspace, islower, isupper, istitle, isnumeric, isdecimal
###################################################################################################
doc = 'Check whether all characters in each string in `column` are alphanumeric'
isalnum = _generate_basic_str_postprocess('isalnum', doc)

doc = 'Check whether all characters in each string in `column` are alphabetic'
isalpha = _generate_basic_str_postprocess('isalpha', doc)

doc = 'Check whether all characters in each string in `column` are digits'
isdigit = _generate_basic_str_postprocess('isdigit', doc)

doc = 'Check whether all characters in each string in `column` are whitespace'
isspace = _generate_basic_str_postprocess('isspace', doc)

doc = 'Check whether all characters in each string in `column` are lowercase'
islower = _generate_basic_str_postprocess('islower', doc)

doc = 'Check whether all characters in each string in `column` are uppercase'
isupper = _generate_basic_str_postprocess('isupper', doc)

doc = 'Check whether all characters in each string in `column` are titlecase'
istitle = _generate_basic_str_postprocess('istitle', doc)

doc = 'Check whether all characters in each string in `column` are numeric'
isnumeric = _generate_basic_str_postprocess('isnumeric', doc)

doc = 'Check whether all characters in each string in `column` are decimal'
isdecimal = _generate_basic_str_postprocess('isdecimal', doc)


###################################################################################################
#                                        STRIP METHODS
#
# All these functions have the same signature:
# :param df: the dataframe
# :param column: the column
# :param to_strip: (str: None) set of characters to be removed
# :param new_column: the destination column (if not set, `column` will be used)
# :return: the transformed dataframe
###################################################################################################
def _generate_strip_str_postprocess(method_name, docstring):
    def f(df, column, *, to_strip=None, new_column=None):
        method = getattr(df[column].str, method_name)
        new_column = new_column or column
        df.loc[:, new_column] = method(to_strip)
        return df

    f.__name__ = method_name
    f.__doc__ = f'{docstring}\n' \
                f':param df: the dataframe\n' \
                f':param column: the column\n' \
                f':param to_strip: (str: None) set of characters to be removed\n' \
                f':param new_column: the destination column (if not set, `column` will be used)\n' \
                f':return: the transformed dataframe'
    return f


doc = 'Strip whitespace (including newlines) from each string in `column` from both sides'
strip = _generate_strip_str_postprocess('strip', doc)

doc = 'Strip whitespace (including newlines) from each string in `column` from left side'
lstrip = _generate_strip_str_postprocess('lstrip', doc)

doc = 'Strip whitespace (including newlines) from each string in `column` from left side'
rstrip = _generate_strip_str_postprocess('rstrip', doc)


###################################################################################################
#                              METHODS with `width` and `fillchar`
#
# All these functions have the same signature:
# :param df: the dataframe
# :param column: the column
# :param width: (int) minimum width
# :param fillchar: (default: \' \') additional character for filling
# :param new_column: the destination column (if not set, `column` will be used)
# :return: the transformed dataframe
###################################################################################################

def _generate_width_str_postprocess(method_name, docstring):
    def f(df, column, *, width, fillchar=' ', new_column=None):
        method = getattr(df[column].str, method_name)
        new_column = new_column or column
        df.loc[:, new_column] = method(width, fillchar=fillchar)
        return df

    f.__name__ = method_name
    f.__doc__ = f'{docstring}\n' \
                f':param df: the dataframe\n' \
                f':param column: the column\n' \
                f':param width (int): minimum width\n' \
                f':param fillchar: (default: \' \') additional character for filling\n' \
                f':param new_column: the destination column (if not set, `column` will be used)\n' \
                f':return: the transformed dataframe'
    return f


doc = 'Filling left and right side of strings in `column` with an additional character'
center = _generate_width_str_postprocess('center', doc)

doc = 'Filling right side of strings in `column` with an additional character'
ljust = _generate_width_str_postprocess('ljust', doc)

doc = 'Filling left side of strings in `column` with an additional character'
rjust = _generate_width_str_postprocess('rjust', doc)


###################################################################################################
#                                        SPLIT METHODS
#
# All these functions have the same signature:
# :param df: the dataframe
# :param column: the column
# :param new_column: the destination column (if not set, `column` will be used)
# :param sep: (default: \' \') string or regular expression to split on
# :param limit: (default: None) limit number of splits in output
# :return: the transformed dataframe
###################################################################################################
def _generate_split_str_postprocess(method_name, docstring):
    def f(df, column, *, new_columns, sep=' ', limit=None):
        method = getattr(df[column].str, method_name)
        df_split = method(pat=sep, n=limit, expand=True)
        nb_cols = df_split.shape[1]
        df[new_columns[:nb_cols]] = df_split
        return df

    f.__name__ = method_name
    f.__doc__ = f'{docstring}\n' \
                f':param df: the dataframe\n' \
                f':param column: the column\n' \
                f':param new_column: the destination column (if not set, `column` will be used)\n' \
                f':param sep: (default: \' \') string or regular expression to split on\n' \
                f':param limit: (default: None) limit number of splits in output\n' \
                f':return: the transformed dataframe'
    return f


doc = 'Split each string in the caller’s values by given pattern, propagating NaN values'
split = _generate_split_str_postprocess('split', doc)

doc = 'Split each string `column` by the given delimiter string, ' \
      'starting at the end of the string and working to the front'
rsplit = _generate_split_str_postprocess('rsplit', doc)


###################################################################################################
#                                        PARTITION METHODS
#
# All these functions have the same signature:
# :param df: the dataframe
# :param column: the column
# :param new_column: the destination column (if not set, `column` will be used)
# :param sep: (default: \' \') string or regular expression to split on
# :return: the transformed dataframe
###################################################################################################
def _generate_partition_str_postprocess(method_name, docstring):
    def f(df, column, *, new_columns, sep=' '):
        if len(new_columns) != 3:
            raise ValueError('`new_columns` must have 3 columns exactly')
        method = getattr(df[column].str, method_name)
        df[new_columns] = method(sep)
        return df

    f.__name__ = method_name
    f.__doc__ = f'{docstring}\n' \
                f':param df: the dataframe\n' \
                f':param column: the column\n' \
                f':param new_column: the destination column (if not set, `column` will be used)\n' \
                f':param sep: (default: \' \') string or regular expression to split on\n' \
                f':return: the transformed dataframe'
    return f


doc = 'Split the string at the first occurrence of sep, and return 3 elements containing ' \
      'the part before the separator, the separator itself, and the part after the separator. ' \
      'If the separator is not found, return 3 elements containing the string itself, ' \
      'followed by two empty strings.'
partition = _generate_partition_str_postprocess('partition', doc)

doc = 'Split the string at the last occurrence of sep, and return 3 elements containing ' \
      'the part before the separator, the separator itself, and the part after the separator. ' \
      'If the separator is not found, return 3 elements containing two empty strings, ' \
      'followed by the string itself.'
rpartition = _generate_partition_str_postprocess('rpartition', doc)


###################################################################################################
#                                   INDEX AND FIND METHODS
#
# All these functions have the same signature:
# :param df: the dataframe
# :param column: the column
# :param new_column: the destination column (if not set, `column` will be used)
# :param sub: substring being searched
# :param start: (default: 0) left edge index
# :param end: (default: None) right edge index
# :return: the transformed dataframe
###################################################################################################
def _generate_find_str_postprocess(method_name, docstring):
    def f(df, column, *, sub, start=0, end=None, new_column=None):
        method = getattr(df[column].str, method_name)
        new_column = new_column or column
        df.loc[:, new_column] = method(sub, start, end)
        return df

    f.__name__ = method_name
    f.__doc__ = f'{docstring}\n' \
                f':param df: the dataframe\n' \
                f':param column: the column\n' \
                f':param new_column: the destination column (if not set, `column` will be used)\n' \
                f':param sub: substring being searched\n' \
                f':param start: (default: 0) left edge index\n' \
                f':param end: (default: None) right edge index\n' \
                f':return: the transformed dataframe'
    return f


doc = 'Return lowest indexes in each strings in `column` where the substring ' \
      'is fully contained between [start:end]. Return -1 on failure'
find = _generate_find_str_postprocess('find', doc)

doc = 'Return highest indexes in each strings in `column` where the substring ' \
      'is fully contained between [start:end]. Return -1 on failure'
rfind = _generate_find_str_postprocess('rfind', doc)

doc = 'Return lowest indexes in each strings where the substring is fully contained ' \
      'between [start:end]. This is the same as str.find except instead of returning -1, ' \
      'it raises a ValueError when the substring is not found'
index = _generate_find_str_postprocess('index', doc)

doc = 'Return highest indexes in each strings where the substring is fully contained ' \
      'between [start:end]. This is the same as str.find except instead of returning -1, ' \
      'it raises a ValueError when the substring is not found'
rindex = _generate_find_str_postprocess('rindex', doc)


###################################################################################################
#                                  STARTSWITH/ENDSWITH METHODS
#
# All these functions have the same signature:
# :param df: the dataframe
# :param column: the column
# :param new_column: the destination column (if not set, `column` will be used)
# :param pat: character sequence
# :param na: (default: NaN) object shown if element tested is not a string
# :return: the transformed dataframe
###################################################################################################
def _generate_with_str_postprocess(method_name, docstring):
    def f(df, column, *, pat, na=np.nan, new_column=None):
        method = getattr(df[column].str, method_name)
        new_column = new_column or column
        df.loc[:, new_column] = method(pat, na=na)
        return df

    f.__name__ = method_name
    f.__doc__ = f'{docstring}\n' \
                f':param df: the dataframe\n' \
                f':param column: the column\n' \
                f':param new_column: the destination column (if not set, `column` will be used)\n' \
                f':param pat: character sequence\n' \
                f':param na: (default: NaN) object shown if element tested is not a string\n' \
                f':return: the transformed dataframe'
    return f


doc = 'Test if the start of each string element matches a pattern.'
startswith = _generate_with_str_postprocess('startswith', doc)

doc = 'Test if the end of each string element matches a pattern.'
endswith = _generate_with_str_postprocess('endswith', doc)


###################################################################################################
#                                        OTHER METHODS
###################################################################################################
def concat(df, *, columns, new_column, sep=None):
    """
    Concatenate `columns` element-wise
    :param df: the dataframe
    :param columns: list of columns to concatenate
    :param new_column: the destination column (if not set, `column` will be used)
    :param sep: (default: None) the separator
    :return: the transformed dataframe
    """
    if len(columns) < 2:
        raise ValueError('The `columns` parameter needs to have at least 2 columns')
    first_col, *other_cols = columns
    df.loc[:, new_column] = df[first_col].astype(str).str.cat(df[other_cols].astype(str), sep=sep)
    return df


def contains(df, column, *, pat, new_column=None, case=True, na=None, regex=True):
    """
    Test if pattern or regex is contained within strings of `column`
    :param df: the dataframe
    :param column: the column
    :param pat: (str) character sequence or regular expression.
    :param new_column: the destination column (if not set, `column` will be used)
    :param case: (bool) if True, case sensitive.
    :param na: fill value for missing values.
    :param regex: (bool) default True
    :return: the transformed dataframe
    """
    new_column = new_column or column
    df.loc[:, new_column] = df[column].str.contains(pat, case=case, na=na, regex=regex)
    return df


def repeat(df, column, *, times, new_column=None):
    """
    Duplicate each string in `column` by indicated number of time
    :param df: the dataframe
    :param column: the column
    :param times: (int) times to repeat the string
    :param new_column: the destination column (if not set, `column` will be used)
    :return: the transformed dataframe
    """
    new_column = new_column or column
    df.loc[:, new_column] = df[column].str.repeat(times)
    return df


def replace_pattern(df, column, *, pat, repl, new_column=None, case=True, regex=True):
    """
    Replace occurrences of pattern/regex in ``column` with some other string
    :param df: the dataframe
    :param column: the column
    :param pat: (str) character sequence or regular expression.
    :param repl: (str) replacement string
    :param new_column: the destination column (if not set, `column` will be used)
    :param case: (bool) if True, case sensitive.
    :param regex: (bool) default True
    :return: the transformed dataframe
    """
    new_column = new_column or column
    df.loc[:, new_column] = df[column].str.replace(pat, repl, case=case, regex=regex)
    return df



import pandas as pd


def two_values_melt(df, first_value_vars, second_value_vars, var_name,
                    value_name):
    """
    First, build two DataFrames from the original one: one to compute a melt
    for the value, another one to compute a melt for the evolution. Second,
    merge these two DataFrames. The idea is to go from something like this:

    | ... | <some1> | <some2> | <some1_evol> | <some2_evol> |
    | ... | <val1>  | <val2>  | <evol1>      | <evol2>      |

    to something like that:

    | ... | variable  | value  | evolution
    | ... | --------- | ------ | ---------
    | ... |  <some1>  | <val1> | <evol1>
    | ... |  <some2>  | <val2> | <evol2>

    Args:
        df (DataFrame): DataFrame to process
        first_value_vars (list): value_vars of a pandas melt, for the first
            value columns of the DataFrame
        second_value_vars (list): value_vars of a pandas melt, for the second
            value columns of the DataFrame
        var_name (str): var_names of a pandas melt
        value_name (str): value_name of a pandas melt

    Notes:
        In tests/app/fixtures, you will find example files for the input and
        output data (respectively two_values_melt_in.csv and
        two_values_melt_out.csv)

    Returns:
        DataFrame: molted DataFrame with two value (value and evolution for
            example) columns

    """
    value_name_first = value_name + '_first'
    value_name_second = value_name + '_second'

    # Melt on the first value columns
    melt_first_value = pd.melt(df,
                               id_vars=[col for col in list(df) if
                                        col not in first_value_vars],
                               value_vars=first_value_vars,
                               var_name=var_name,
                               value_name=value_name_first)
    melt_first_value.drop(second_value_vars, axis=1, inplace=True)

    # Melt on the second value columns
    melt_second_value = pd.melt(df,
                                id_vars=[col for col in list(df) if
                                         col not in second_value_vars],
                                value_vars=second_value_vars,
                                var_name=var_name,
                                value_name=value_name_second)

    # Since there are two value columns, there is no need to keep the
    # second_value_vars names. And it will make things easier for the merge.
    normalize_types = {k: v for k, v in zip(second_value_vars, first_value_vars)}
    melt_second_value.replace(normalize_types, inplace=True)
    melt_second_value.drop(first_value_vars, axis=1, inplace=True)

    on_cols = list(melt_first_value)
    on_cols.remove(value_name_first)
    return pd.merge(melt_first_value, melt_second_value, on=on_cols, how='outer')



from numpy import nan


def fillna(df, column, value=None,  column_value=None):
    """
    Can fill NaN values from a column - you can fill with column or with value
    Args:
        df
        column
        value
        column_value
    """
    if column not in df.columns:
        df[column] = nan

    if value is not None and column_value is not None:
        raise ValueError('You cannot set both the value parameter and column_value parameter')

    if value is not None:
        df[column] = df[column].fillna(value)

    if column_value is not None:
        if column_value not in df.columns:
            raise ValueError(f'"{column_value}" is not a valid column name')
        df[column] = df[column].fillna(df[column_value])

    return df
